@@ -5,7 +5,7 @@ To turn on verbose logging
 import logging
 logging.getLogger('calcbench.api_client').setLevel(logging.INFO)
 """
-__version__ = "3.0.1"
+__version__ = "3.0.3"
 from .api_client import (
     normalized_data,
     normalized_dataframe,
@@ -39,7 +39,7 @@ from .api_client import (
     raw_data,
     raw_data_raw,
     face_statement,
-    CompanyIdentifierScheme
+    CompanyIdentifierScheme,
 )
 
-from .listener import handle_filings
\ No newline at end of file
+from .listener import handle_filings
