@@ -6,15 +6,15 @@ Created on Mar 14, 2015
 @contact: andrew@calcbench.com
 """
 
-from __future__ import print_function
 import os
+from typing import Dict, Generator, Iterable, Literal, Optional, Sequence, Union
 import requests
 import json
 import warnings
-from datetime import datetime
+from datetime import date, datetime
 from functools import wraps
 import logging
-from enum import Enum
+from enum import Enum, IntEnum
 
 logger = logging.getLogger(__name__)
 
@@ -36,6 +36,7 @@ try:
 except ImportError:
     pass
 
+
 _SESSION_STUFF = {
     "calcbench_user_name": os.environ.get("CALCBENCH_USERNAME"),
     "calcbench_password": os.environ.get("CALCBENCH_PASSWORD"),
@@ -46,6 +47,7 @@ _SESSION_STUFF = {
     "session": None,
     "timeout": 60 * 20,  # twenty minute content request timeout, by default
     "enable_backoff": False,
+    "proxies": None,
 }
 
 
@@ -56,6 +58,7 @@ def _calcbench_session():
         password = _SESSION_STUFF.get("calcbench_password")
         if not (user_name and password):
             import getpass
+
             user_name = input(
                 'Calcbench username/email. Set the "calcbench_user_name" environment variable or call "set_credentials" to avoid this prompt::'
             )
@@ -93,6 +96,16 @@ def _rig_for_testing(domain="localhost:444", suppress_http_warnings=True):
         requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
 
 
+PeriodType = Literal["annual", "quarterly"]
+
+
+CentralIndexKey = Union[str, int]
+Ticker = str
+CalcbenchCompanyIdentifier = int
+CompanyIdentifier = Union[Ticker, CentralIndexKey, CalcbenchCompanyIdentifier]
+CompanyIdentifiers = Sequence[CompanyIdentifier]
+
+
 def _add_backoff(f):
     @wraps(f)
     def wrapper(*args, **kwargs):
@@ -188,7 +201,7 @@ def enable_backoff(backoff_on=True, giveup=lambda e: False):
     _SESSION_STUFF["backoff_giveup"] = giveup
 
 
-def set_proxies(proxies):
+def set_proxies(proxies: Dict[str, str]):
     """
         Set proxies used for requests.  See https://requests.readthedocs.io/en/master/user/advanced/#proxies
 
@@ -196,54 +209,66 @@ def set_proxies(proxies):
     _SESSION_STUFF["proxies"] = proxies
 
 
-class CompanyIdentifierScheme(Enum):
+class CompanyIdentifierScheme(str, Enum):
     Ticker = "ticker"
     CentralIndexKey = "CIK"
 
 
+class Period(IntEnum):
+    Annual = 0
+    Q1 = 1
+    Q2 = 2
+    Q3 = 3
+    Q4 = 4
+
+
+PeriodArgument = Optional[Union[Period, Literal[0, 1, 2, 3, 4]]]
+
+
 def normalized_dataframe(
-    company_identifiers=[],
-    metrics=[],
-    start_year=None,
-    start_period=None,
-    end_year=None,
-    end_period=None,
-    entire_universe=False,
-    filing_accession_number=None,
-    point_in_time=False,
-    year=None,
-    period=None,
-    all_history=False,
-    period_type=None,
-    trace_hyperlinks=False,
-    use_fiscal_period=False,
-    company_identifier_scheme=CompanyIdentifierScheme.Ticker,
-):
+    company_identifiers: CompanyIdentifiers = [],
+    metrics: Sequence[str] = [],
+    start_year: int = None,
+    start_period: PeriodArgument = None,
+    end_year: int = None,
+    end_period: PeriodArgument = None,
+    entire_universe: bool = False,
+    filing_accession_number: int = None,
+    point_in_time: bool = False,
+    year: int = None,
+    period: PeriodArgument = None,
+    all_history: bool = False,
+    period_type: PeriodType = None,
+    trace_hyperlinks: bool = False,
+    use_fiscal_period: bool = False,
+    company_identifier_scheme: CompanyIdentifierScheme = CompanyIdentifierScheme.Ticker,
+) -> "pd.DataFrame":
     """Standardized Data.
     
     Metrics are standardized by economic concept and time period.  
     
     The data behind the multi-company page, https://www.calcbench.com/multi.
+
+    Example https://github.com/calcbench/notebooks/blob/master/python_client_api_demo.ipynb
     
-    :param sequence company_identifiers: Tickers/CIK codes. eg. ['msft', 'goog', 'appl', '0000066740']
-    :param sequence metrics: Standardized metrics.  Full list @ https://www.calcbench.com/home/standardizedmetrics eg. ['revenue', 'accountsreceivable']
-    :param int start_year: first year of data
-    :param int start_period: first quarter to get, for annual data pass 0, for quarters pass 1, 2, 3, 4
-    :param int end_year: last year of data
-    :param int end_period: last_quarter to get, for annual data pass 0, for quarters pass 1, 2, 3, 4        
-    :param bool entire_universe: Get data for all companies, this can take a while, talk to Calcbench before you do this in production.
-    :param int accession_id: Filing Accession ID from the SEC's Edgar system.
-    :param int year: Get data for a single year, defaults to annual data.
-    :param str period_type: Either "annual" or "quarterly".
+    :param company_identifiers: Tickers/CIK codes. eg. ['msft', 'goog', 'appl', '0000066740']
+    :param metrics: Standardized metrics.  Full list @ https://www.calcbench.com/home/standardizedmetrics eg. ['revenue', 'accountsreceivable']
+    :param start_year: first year of data
+    :param start_period: first quarter to get, for annual data pass 0, for quarters pass 1, 2, 3, 4
+    :param end_year: last year of data
+    :param end_period: last_quarter to get, for annual data pass 0, for quarters pass 1, 2, 3, 4        
+    :param entire_universe: Get data for all companies, this can take a while, talk to Calcbench before you do this in production.
+    :param filing_accession_number: Filing Accession ID from the SEC's Edgar system.
+    :param year: Get data for a single year, defaults to annual data.
+    :param period_type: Either "annual" or "quarterly".
     :return: Dataframe with the periods as the index and columns indexed by metric and ticker
-    :rtype: Dataframe
 
     Usage::
 
       >>> calcbench.standardized_data(company_identifiers=['msft', 'goog'], metrics=['revenue', 'assets'], all_history=True, period_type='annual')
 
     """
-    company_identifier_scheme = CompanyIdentifierScheme(company_identifier_scheme)
+
     if all_history and not period_type:
         raise ValueError("For all history you must specify a period_type")
     data = normalized_raw(
@@ -292,7 +317,7 @@ def normalized_dataframe(
         warnings.warn("Did not find metrics {0}".format(missing_metrics))
     data = pd.DataFrame(data)
     data.set_index(
-        keys=[company_identifier_scheme.value, "metric", "period"], inplace=True
+        keys=[f"{company_identifier_scheme}", "metric", "period"], inplace=True
     )
     try:
         data = data.unstack("metric")["value"]
@@ -312,7 +337,7 @@ def normalized_dataframe(
 
     for missing_metric in missing_metrics:
         data[missing_metric] = np.nan  # We want columns for every requested metric.
-    data = data.unstack(company_identifier_scheme.value)
+    data = data.unstack(f"{company_identifier_scheme}")
     return data
 
 
@@ -345,23 +370,25 @@ standardized_data = normalized_dataframe  # Now it's called standardized data
 
 
 def normalized_raw(
-    company_identifiers=[],
-    metrics=[],  # type str[] Full list of metrics is @ https://www.calcbench.com/home/standardizedmetrics
-    start_year=None,
-    start_period=None,
-    end_year=None,
-    end_period=None,
+    company_identifiers: CompanyIdentifiers = [],
+    metrics: Iterable[
+        str
+    ] = [],  # type str[] Full list of metrics is @ https://www.calcbench.com/home/standardizedmetrics
+    start_year: int = None,
+    start_period: PeriodArgument = None,
+    end_year: int = None,
+    end_period: PeriodArgument = None,
     entire_universe=False,
     filing_accession_number=None,
-    point_in_time=False,
-    include_trace=False,
-    update_date=None,
+    point_in_time: bool = False,
+    include_trace: bool = False,
+    update_date: date = None,
     all_history=False,
-    year=None,
-    period=None,
-    period_type=None,
-    include_preliminary=False,
-    use_fiscal_period=False,
+    year: int = None,
+    period: PeriodArgument = None,
+    period_type: PeriodType = None,
+    include_preliminary: bool = False,
+    use_fiscal_period: bool = False,
 ):
     """
     Standardized data.
@@ -493,36 +520,36 @@ def normalized_raw(
 
 
 def point_in_time(
-    company_identifiers=[],
-    all_footnotes=False,
-    update_date=None,
-    metrics=[],
-    all_history=False,
-    entire_universe=False,
-    start_year=None,
-    start_period=None,
-    end_year=None,
-    end_period=None,
-    period_type=None,
-    use_fiscal_period=False,
-    include_preliminary=False,
-    all_face=False,
-    include_xbrl=True,
-    accession_id=None,
-    include_trace=False,
-):
+    company_identifiers: CompanyIdentifiers = [],
+    all_footnotes: bool = False,
+    update_date: date = None,
+    metrics: Iterable[str] = [],
+    all_history: bool = False,
+    entire_universe: bool = False,
+    start_year: int = None,
+    start_period: PeriodArgument = None,
+    end_year: int = None,
+    end_period: PeriodArgument = None,
+    period_type: PeriodType = None,
+    use_fiscal_period: bool = False,
+    include_preliminary: bool = False,
+    all_face: bool = False,
+    include_xbrl: bool = True,
+    accession_id: int = None,
+    include_trace: bool = False,
+) -> "pd.DataFrame":
     """Point-in-Time Data
 
     Standardized data with a timestamp when it was published by Calcbench
     
-    :param date update_date: The date on which the data was received, this does not work prior to ~2016, use all_history to get historical data then use update_date to get updates.
-    :param int accession_id: Unique identifier for the filing for which to recieve data.  Pass this to recieve data for one filing.  Same as filing_id in filings objects
-    :param bool all_face: Retrieve all of the points from the face financials, income/balance/statement of cash flows
-    :param bool all_footnotes: Retrive all of the points from the footnotes to the financials
+    :param update_date: The date on which the data was received, this does not work prior to ~2016, use all_history to get historical data then use update_date to get updates.
+    :param accession_id: Unique identifier for the filing for which to recieve data.  Pass this to recieve data for one filing.  Same as filing_id in filings objects
+    :param all_face: Retrieve all of the points from the face financials, income/balance/statement of cash flows
+    :param all_footnotes: Retrive all of the points from the footnotes to the financials
+    :param include_preliminary: Include facts from non-XBRL earnings press-releases and 8-Ks.
+    :param include_xbrl: Include facts from XBRL 10-K/Qs.
     :return: DataFrame of facts
 
-    :rtype: pandas.DataFrame
-
     Columns:
 
 
@@ -632,7 +659,7 @@ def mapped_raw(
     include_xbrl=True,
     accession_id=None,
     include_trace=False,
-) -> pd.DataFrame:
+) -> "pd.DataFrame":
     if update_date:
         warnings.warn(
             "Request updates by accession_id rather than update date",
@@ -778,28 +805,28 @@ def dimensional_raw(
 
 
 def document_dataframe(
-    company_identifiers=[],
-    disclosure_names=[],
-    all_history=False,
-    year=None,
-    period=None,
-    progress_bar=None,
-    period_type=None,
-    identifier_key="ticker",
-):
+    company_identifiers: CompanyIdentifiers = [],
+    disclosure_names: Iterable[str] = [],
+    all_history: bool = False,
+    year: int = None,
+    period: PeriodArgument = None,
+    progress_bar: "tqdm.std.tqdm" = None,
+    period_type: PeriodType = None,
+    identifier_key: Literal["ticker", "CIK"] = "ticker",
+    block_tag_names: Sequence[str] = [],
+    use_fiscal_period=False,
+) -> "pd.DataFrame":
     """Disclosures/Footnotes in a DataFrame
 
-    :param list(str) company_identifiers: list of tickers or CIK codes
-    :param list(str) disclosure_names: The sections to retrieve, see the full list @ https://www.calcbench.com/disclosure_list.  You cannot request XBRL and non-XBRL sections in the same request.  eg.  ['Management's Discussion And Analysis', 'Risk Factors']
-    :param bool all_history: Search all time periods
-    :param int year: The year to search
-    :param int period: period of data to get.  0 for annual data, 1, 2, 3, 4 for quarterly data.
-    :param str period_type: "quarterly" or "annual", only applicable when other period data not supplied.  Use "annual" to only search end-of-year documents.
+    :param company_identifiers: list of tickers or CIK codes
+    :param disclosure_names: The sections to retrieve, see the full list @ https://www.calcbench.com/disclosure_list.  You cannot request XBRL and non-XBRL sections in the same request.  eg.  ['Management's Discussion And Analysis', 'Risk Factors']
+    :param all_history: Search all time periods
+    :param year: The year to search
+    :param period: period of data to get.  0 for annual data, 1, 2, 3, 4 for quarterly data.
+    :param period_type: "quarterly" or "annual", only applicable when other period data not supplied.  Use "annual" to only search end-of-year documents.
     :param tqdm.tqdm progress_bar: Pass a tqdm progress bar to keep an eye on things.
-    :param string identifier_key: "ticker" or "CIK", how to index the returned DataFrame.
+    :param identifier_key: "ticker" or "CIK", how to index the returned DataFrame.
     :return: A DataFrame indexed by document name -> company identifier.
-    :rtype: pandas.DataFrame
-
 
     Usage::      
 
@@ -808,27 +835,44 @@ def document_dataframe(
       >>> word_counts = data.applymap(lambda document: document and len(document.get_contents_text().split()))
       
     """
-
-    docs = list(
-        document_search(
-            company_identifiers=company_identifiers,
-            disclosure_names=disclosure_names,
-            all_history=all_history,
-            use_fiscal_period=True,
-            progress_bar=progress_bar,
-            year=year,
-            period=period,
-            period_type=period_type,
+    if block_tag_names:
+        docs = []
+        for block_tag_name in block_tag_names:
+            docs.extend(
+                document_search(
+                    company_identifiers=company_identifiers,
+                    block_tag_name=block_tag_name,
+                    all_history=all_history,
+                    use_fiscal_period=use_fiscal_period,
+                    progress_bar=progress_bar,
+                    year=year,
+                    period=period,
+                    period_type=period_type,
+                )
+            )
+    else:
+        docs = list(
+            document_search(
+                company_identifiers=company_identifiers,
+                disclosure_names=disclosure_names,
+                all_history=all_history,
+                use_fiscal_period=True,
+                progress_bar=progress_bar,
+                year=year,
+                period=period,
+                period_type=period_type,
+            )
         )
-    )
     period_map = {"1Q": 1, "2Q": 2, "3Q": 3, "Y": 4}
     for doc in docs:
-        period_year = doc["fiscal_year"]
+        period_year = doc["fiscal_year" if use_fiscal_period else "calendar_year"]
         if period in ("Y", 0) or period_type == "annual":
             p = pd.Period(year=period_year, freq="a")
         else:
             try:
-                quarter = period_map[doc["fiscal_period"]]
+                quarter = period_map[
+                    doc["fiscal_period" if use_fiscal_period else "calendar_period"]
+                ]
             except KeyError:
                 # This happens for non-XBRL companies
                 logger.info("Strange period for {ticker}".format(**doc))
@@ -846,44 +890,68 @@ def document_dataframe(
     return data
 
 
+class DocumentSearchResults(dict):
+    """
+        Represents an disclosure.
+    """
+
+    def get_contents(self):
+        """
+            Content of the document, with the filers HTML
+        """
+        if self.get("network_id"):
+            return _document_contents_by_network_id(self["network_id"])
+        else:
+            return document_contents(
+                blob_id=self["blob_id"], SEC_ID=self["sec_filing_id"]
+            )
+
+    def get_contents_text(self):
+        """Contents of the HTML of the document"""
+        return "".join(BeautifulSoup(self.get_contents(), "html.parser").strings)
+
+    @property
+    def date_reported(self) -> datetime:
+        """Time (EST) the document was available from Calcbench"""
+        return _try_parse_timestamp(self["date_reported"])
+
+
 def document_search(
-    company_identifiers=None,
-    full_text_search_term=None,
-    year=None,
-    period=0,
-    period_type=None,
-    document_type=None,
-    block_tag_name=None,
-    entire_universe=False,
-    use_fiscal_period=False,
-    document_name=None,
-    all_history=False,
-    updated_from=None,
-    batch_size=100,
-    sub_divide=False,
-    all_documents=False,
-    disclosure_names=[],
-    progress_bar=None,
-    accession_id=None,
-):
+    company_identifiers: CompanyIdentifiers = None,
+    full_text_search_term: str = None,
+    year: int = None,
+    period: PeriodArgument = Period.Annual,
+    period_type: PeriodType = None,
+    document_type: str = None,
+    block_tag_name: str = None,
+    entire_universe: bool = False,
+    use_fiscal_period: bool = False,
+    document_name: bool = None,
+    all_history: bool = False,
+    updated_from: date = None,
+    batch_size: int = 100,
+    sub_divide: bool = False,
+    all_documents: bool = False,
+    disclosure_names: Sequence[str] = [],
+    progress_bar: "tqdm.std.tqdm" = None,
+    accession_id: int = None,
+) -> Generator[DocumentSearchResults, None, None]:
     """
     Footnotes and other text
     
     Search for footnotes and other sections of 10-K, see https://www.calcbench.com/footnote.
     
-    :param list(str) company_identifiers: list of tickers or CIK codes
-    :param int year: Year to get data for
-    :param int period: period of data to get.  0 for annual data, 1, 2, 3, 4 for quarterly data.
-    :param str period_type: "quarterly" or "annual", only applicable when other period data not supplied.  Use "annual" to only search end-of-year documents.
-    :param list(str) document_names:  The sections to retrieve, see the full list @ https://www.calcbench.com/disclosure_list.  You cannot request XBRL and non-XBRL sections in the same request.  eg.  ['Management's Discussion And Analysis', 'Risk Factors'] 
-    :param bool all_history: Search all time periods
-    :param datetime.date updated_from: date, include filings from this date and after.
-    :param bool sub_divide: return the document split into sections based on headers.
-    :param bool all_documents: all of the documents for a single company/period.
-    :param bool entire_universe: Search all companies
-    :param tqdm.tqdm progress_bar: Pass a tqdm progress bar to keep an eye on things.
-    :return: A generator of DocumentSearchResults
-    :rtype: generator(DocumentSearchResults)
+    :param company_identifiers: list of tickers or CIK codes
+    :param year: Year to get data for
+    :param period: period of data to get.  0 for annual data, 1, 2, 3, 4 for quarterly data.
+    :param period_type: only applicable when other period data not supplied.  Use "annual" to only search end-of-year documents.
+    :param document_names:  The sections to retrieve, see the full list @ https://www.calcbench.com/disclosure_list.  You cannot request XBRL and non-XBRL sections in the same request.  eg.  ['Management's Discussion And Analysis', 'Risk Factors'] 
+    :param all_history: Search all time periods
+    :param updated_from: date, include filings from this date and after.
+    :param sub_divide: return the document split into sections based on headers.
+    :param all_documents: all of the documents for a single company/period.
+    :param entire_universe: Search all companies
+    :param progress_bar: Pass a tqdm progress bar to keep an eye on things.
 
     Usage::
     
@@ -914,7 +982,9 @@ def document_search(
         if not year:
             raise ValueError("Need to specify year or all all_history")
         period_type = (
-            period_type or "annual" if period in (0, "Y", "y") else "quarterly"
+            period_type or PeriodType.Annual
+            if period in (0, "Y", "y")
+            else PeriodType.Quarterly
         )
     payload = {
         "companiesParameters": {"entireUniverse": entire_universe},
@@ -964,26 +1034,6 @@ def _document_search_results(payload, progress_bar=None):
     payload["pageParameters"]["startOffset"] = None
 
 
-class DocumentSearchResults(dict):
-    def get_contents(self):
-        """Content of the document, with the filers HTML"""
-        if self.get("network_id"):
-            return _document_contents_by_network_id(self["network_id"])
-        else:
-            return document_contents(
-                blob_id=self["blob_id"], SEC_ID=self["sec_filing_id"]
-            )
-
-    def get_contents_text(self):
-        """Contents of the HTML of the document"""
-        return "".join(BeautifulSoup(self.get_contents(), "html.parser").strings)
-
-    @property
-    def date_reported(self):
-        """Time (EST) the document was available from Calcbench"""
-        return _try_parse_timestamp(self["date_reported"])
-
-
 def _try_parse_timestamp(timestamp):
     """
     We did not always have milliseconds
@@ -995,13 +1045,13 @@ def _try_parse_timestamp(timestamp):
         return datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")
 
 
-def document_contents(blob_id, SEC_ID, SEC_URL=None):
+def document_contents(blob_id, SEC_ID, SEC_URL=None) -> str:
     payload = {"blobid": blob_id, "secid": SEC_ID, "url": SEC_URL}
     json = _json_GET("query/disclosureBySECLink", payload)
     return json["blobs"][0]
 
 
-def _document_contents_by_network_id(network_id):
+def _document_contents_by_network_id(network_id) -> str:
     payload = {"nid": network_id}
     json = _json_GET("query/disclosureByNetworkIDOBJ", payload)
     blobs = json["blobs"]
@@ -1156,6 +1206,10 @@ def business_combinations(company_identifiers):
     return _json_POST("businessCombinations", payload)
 
 
+class Filing(dict):
+    pass
+
+
 def filings(
     company_identifiers=[],
     entire_universe=False,
@@ -1164,7 +1218,7 @@ def filings(
     start_date=None,
     end_date=None,
     filing_types=[],
-):
+) -> Iterable[Filing]:
     """SEC filings
 
     https://www.calcbench.com/filings
@@ -1204,7 +1258,7 @@ def filings(
         yield _cast_filing_fields(filing)
 
 
-def _cast_filing_fields(filing):
+def _cast_filing_fields(filing: Filing) -> Filing:
     for date_field in (
         "calcbench_finished_load",
         "calcbench_accepted",
